/**
 * Sanitizes objects and records according to the corresponding schema.
 * 'Sanitize' in this context means to recover JS data types like `Date` from plain
 * string properties from the meta-information in the resource schema.
 *
 * @param {Object||Object[]} obj
 * @param {Schema} schema
 */
module.exports = function sanitize(obj, schema) {
  if (!schema.properties && !schema.allProperties)
    return;

  var properties = schema.allProperties || schema.properties;
  if (!Array.isArray(obj))
    Object.keys(obj).forEach(function(key) {
      var prop = properties[key];
      if (!prop)
        return;

      if ((prop.type === 'object' || prop.type === 'array')
          && typeof obj[key] == 'object') {
        sanitize(obj[key], prop);
      }
      // sanitize date-time strings
      else if ((prop.type === 'string'
          || (prop['$ref'] && prop['$ref'].indexOf('string') > 0))
          && prop.format === 'date-time'
          && typeof obj[key] == 'string') {
        obj[key] = new Date(obj[key]);
      }
      // sanitize integers
      else if (prop.type === 'integer' && typeof obj[key] == 'string') {
        obj[key] = parseInt(obj[key]);
      }
      // sanitize numbers
      else if (prop.type === 'number' && typeof obj[key] == 'string') {
        obj[key] = parseFloat(obj[key]);
      }
    });
  else if (schema.items)
    for (var i = 0; i < obj.length; i++) {
      sanitize(obj[i], schema.items);
    }
};
